import {PostToConnectionCommand} from '@aws-sdk/client-apigatewaymanagementapi';
import {AnyTRPCRouter, callTRPCProcedure, inferRouterContext} from '@trpc/server';
import type {APIGatewayResult} from '@trpc/server/adapters/aws-lambda';
import type {Observable} from '@trpc/server/observable';
import {APIGatewayProxyEvent} from 'aws-lambda';
import {getApiGatewayManagementApiClient} from '../getApiGatewayManagementApiClient';
import {getConnection} from '../getConnection';
import {TrpcSubscriptionRequest} from './TrpcSubscriptionRequest';
import {TrpcSubscriptionResponse} from './TrpcSubscriptionResponse';

export const awsLambdaTrpcSubscriptionRequestHandler = <Router extends AnyTRPCRouter>(opts: {
  router: Router
  createContext?: (event: APIGatewayProxyEvent) => Promise<inferRouterContext<Router>>
}) => {
  const {router, createContext} = opts;

  return async (event: APIGatewayProxyEvent): Promise<APIGatewayResult> => {
    const connection = getConnection(event);

    if (!event.body || event.body === '[]')
      return {statusCode: 200, body: ''};

    const request = (JSON.parse(event.body) as TrpcSubscriptionRequest);

    const sendMessage = async (response: TrpcSubscriptionResponse) => {
      await getApiGatewayManagementApiClient(connection.endpoint).send(
          new PostToConnectionCommand({
            ConnectionId: connection.connectionId,
            Data: JSON.stringify(response),
          }),
      );
    };
    const stopConnection = async (): Promise<void> => {
      await sendMessage({
        id: request.id,
        result: {type: 'stopped'},
      });
    };

    try {
      if (request.method === 'subscription.stop') {
        await stopConnection();
      }

      if (request.method === 'subscription') {
        await sendMessage({
          id: request.id,
          result: {type: 'started'},
        });

        const sub = (await callTRPCProcedure({
          procedures: router._def.procedures,
          path: request.params.path,
          type: 'subscription',
          ctx: createContext?.(event) || {},
          getRawInput: async () => request.params.input,
        })) as Observable<unknown, unknown>;

        const messageQueue: TrpcSubscriptionResponse[] = [];
        sub.subscribe({
          next: (data: unknown) => {
            messageQueue.push({
              id: request.id,
              result: {
                type: 'data',
                data: data,
              },
            });
          },
          error: (err) => {
            console.error(err);
            messageQueue.push({
              id: request.id,
              result: {type: 'stopped'},
            });
          },
          complete: () => {
            messageQueue.push({
              id: request.id,
              result: {type: 'stopped'},
            });
          },
        });

        let stopped = false;
        while (!stopped) {
          const message = messageQueue.shift();
          if (!message)
            continue;

          await sendMessage(message);

          stopped = message.result.type === 'stopped';
        }
      }
    } catch (err) {
      console.error(err);
      await stopConnection();
    }

    return {
      statusCode: 200,
      body: '',
    };
  };
};
